import functools
import sys
from tempfile import NamedTemporaryFile
from traceback import format_exception

from botocore.exceptions import ClientError, NoCredentialsError


def store_exception(exception: Exception) -> str:
    """
    Stores the exception in a temporary file and returns its filename
    """

    tracebacks = format_exception(etype=type(exception), value=exception, tb=exception.__traceback__)  # type: [str]

    content = "".join(tracebacks)

    with NamedTemporaryFile(prefix="piu-traceback-", delete=False) as error_file:
        file_name = error_file.name
        error_file.write(content.encode())

    return file_name


def is_credentials_expired_error(e: ClientError) -> bool:
    return e.response["Error"]["Code"] in ["ExpiredToken", "RequestExpired"]


def handle_exceptions(func):
    @functools.wraps(func)
    def wrapper():
        try:
            func()
        except NoCredentialsError as e:
            print(
                'No AWS credentials found. Use the "zaws" command-line tool to get a temporary access key\n'
                "or manually configure either ~/.aws/credentials or AWS_ACCESS_KEY_ID/AWS_SECRET_ACCESS_KEY.",
                file=sys.stderr,
            )
            sys.exit(1)
        except ClientError as e:
            sys.stdout.flush()
            if is_credentials_expired_error(e):
                print(
                    "AWS credentials have expired.\n"
                    'Use the "zaws" command line tool to get a new temporary access key.',
                    file=sys.stderr,
                )
                sys.exit(1)
            else:
                file_name = store_exception(e)
                print("Unknown Error.\n" "Please create an issue with the content of {fn}".format(fn=file_name))
                sys.exit(1)
        except Exception as e:
            # Catch All

            file_name = store_exception(e)
            print("Unknown Error.\n" "Please create an issue with the content of {fn}".format(fn=file_name))
            sys.exit(1)

    return wrapper



import boto3
import click
from collections import namedtuple

Instance = namedtuple("Instance", ["instance_id", "name", "stack_name", "stack_version", "private_ip"])


def _hosted_zones(route53):
    """Enumerates all Route53 hosted zones as tuples of id, name"""

    paginator = route53.get_paginator("list_hosted_zones")
    for result in paginator.paginate():
        for zone in result["HostedZones"]:
            yield zone["Id"], zone["Name"]


def find_odd_host(region):
    """Returns the Odd SSH bastion hostname for the current AWS account and the
       specified region if it exists"""
    if not region:
        return

    route53 = boto3.client("route53", region_name=region)
    for zone_id, zone_name in _hosted_zones(route53):
        candidate_host = "odd-{}.{}".format(region, zone_name)
        result = route53.list_resource_record_sets(
            HostedZoneId=zone_id, MaxItems="1", StartRecordType="A", StartRecordName=candidate_host
        )
        for record in result["ResourceRecordSets"]:
            if record["Type"] == "A" and record["Name"] == candidate_host:
                return record["Name"].rstrip(".")


def list_running_instances(region, filters):
    """Generator that yields Instance records for running EC2 instances matching the
       filter and region"""
    ec2 = boto3.resource("ec2", region_name=region)
    effective_filters = [{"Name": "instance-state-name", "Values": ["running"]}]
    effective_filters.extend(filters)
    for instance in ec2.instances.filter(Filters=effective_filters):
        instance_id = instance.instance_id

        raw_tags = instance.tags or {}
        tags = {tag["Key"]: tag["Value"] for tag in raw_tags}

        yield Instance(
            instance_id, tags.get("Name"), tags.get("StackName"), tags.get("StackVersion"), instance.private_ip_address
        )


def current_region():
    """Returns the current AWS region"""
    session = boto3.session.Session()
    return session.region_name


def validate_region(ctx, param, value):
    """Click validator for the AWS region argument"""
    if value is not None:
        session = boto3.session.Session()
        if value not in session.get_available_regions("cloudformation"):
            raise click.BadParameter("Invalid AWS region '{}'".format(value))

    return value



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
import inspect

import setuptools
from setuptools.command.test import test as TestCommand
from setuptools import setup

if sys.version_info < (3, 4, 0):
    sys.stderr.write("FATAL: STUPS Piu needs to be run with Python 3.4+\n")
    sys.exit(1)

__location__ = os.path.join(os.getcwd(), os.path.dirname(inspect.getfile(inspect.currentframe())))


def read_version(package):
    data = {}
    with open(os.path.join(package, "__init__.py"), "r") as fd:
        exec(fd.read(), data)
    return data["__version__"]


NAME = "stups-piu"
MAIN_PACKAGE = "piu"
VERSION = read_version(MAIN_PACKAGE)
DESCRIPTION = 'Command line client for "even" SSH access granting service'
LICENSE = "Apache License 2.0"
URL = "https://github.com/zalando-stups/piu"
AUTHOR = "Henning Jacobs"
EMAIL = "henning.jacobs@zalando.de"

COVERAGE_XML = True
COVERAGE_HTML = False
JUNIT_XML = True

# Add here all kinds of additional classifiers as defined under
# https://pypi.python.org/pypi?%3Aaction=list_classifiers
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Developers",
    "Intended Audience :: System Administrators",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: POSIX :: Linux",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: Implementation :: CPython",
]

CONSOLE_SCRIPTS = ["piu = piu.cli:main"]


class PyTest(TestCommand):

    user_options = [
        ("cov=", None, "Run coverage"),
        ("cov-xml=", None, "Generate junit xml report"),
        ("cov-html=", None, "Generate junit html report"),
    ]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.cov = None
        self.cov_xml = False
        self.cov_html = False

    def finalize_options(self):
        TestCommand.finalize_options(self)
        if self.cov is not None:
            self.cov = ["--cov", self.cov, "--cov-report", "term-missing"]
            if self.cov_xml:
                self.cov.extend(["--cov-report", "xml"])
            if self.cov_html:
                self.cov.extend(["--cov-report", "html"])

    def run_tests(self):
        try:
            import pytest
        except ImportError:
            raise RuntimeError("py.test is not installed, run: pip install pytest")
        params = {"args": self.test_args}
        if self.cov:
            params["args"] += self.cov
        params["args"] += ["--doctest-modules", MAIN_PACKAGE, "-s"]
        errno = pytest.main(**params)
        sys.exit(errno)


def get_install_requirements(path):
    content = open(os.path.join(__location__, path)).read()
    return [req for req in content.split("\\n") if req != ""]


def read(fname):
    return open(os.path.join(__location__, fname), encoding="utf-8").read()


def setup_package():
    # Assemble additional setup commands
    cmdclass = {}
    cmdclass["test"] = PyTest

    # Some helper variables
    version = VERSION

    install_reqs = get_install_requirements("requirements.txt")

    command_options = {"test": {"test_suite": ("setup.py", "tests"), "cov": ("setup.py", MAIN_PACKAGE)}}
    if COVERAGE_XML:
        command_options["test"]["cov_xml"] = "setup.py", True
    if COVERAGE_HTML:
        command_options["test"]["cov_html"] = "setup.py", True

    setup(
        name=NAME,
        version=version,
        url=URL,
        description=DESCRIPTION,
        author=AUTHOR,
        author_email=EMAIL,
        license=LICENSE,
        keywords="aws account saml login federated shibboleth",
        long_description=read("README.rst"),
        classifiers=CLASSIFIERS,
        test_suite="tests",
        packages=setuptools.find_packages(exclude=["tests", "tests.*"]),
        install_requires=install_reqs,
        setup_requires=["six", "black"],
        cmdclass=cmdclass,
        tests_require=["pytest-cov", "pytest"],
        command_options=command_options,
        entry_points={"console_scripts": CONSOLE_SCRIPTS},
    )


if __name__ == "__main__":
    setup_package()



#!/usr/bin/env python3
"""
Helper script to request access to a certain host.
"""

import boto3
import click
import datetime
import operator
import ipaddress
import json
import os
import subprocess

import dateutil
import requests
import socket
import sys
import time
import yaml
import zign.api
import re

from clickclick import error, AliasedGroup, print_table, OutputFormat, warning
from .error_handling import handle_exceptions

import piu
import piu.utils

COMPATIBLE_AMI_CUTOFF = "2020-02-19T12:00:00.000Z"

try:
    import pyperclip
except ImportError:
    pyperclip = None

CONFIG_DIR_PATH = click.get_app_dir("piu")
CONFIG_FILE_PATH = os.path.join(CONFIG_DIR_PATH, "piu.yaml")

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
DEFAULT_COMMAND = "request-access"

STUPS_CIDR = ipaddress.ip_network("172.31.0.0/16")

STATUS_NAMES = ["REQUESTED", "GRANTED", "DENIED", "FAILED", "EXPIRED", "REVOKED"]

STYLES = {
    "REQUESTED": {"fg": "yellow", "bold": True},
    "GRANTED": {"fg": "green"},
    "DENIED": {"fg": "red", "bold": True},
    "FAILED": {"fg": "red", "bold": True},
    "EXPIRED": {"fg": "yellow", "bold": True},
    "REVOKED": {"fg": "red"},
    "OK": {"fg": "green"},
    "ERROR": {"fg": "red"},
}

TITLES = {"created_time": "Created", "lifetime_minutes": "TTL"}

MAX_COLUMN_WIDTHS = {"reason": 50, "remote_host": 20, "status_reason": 50}

output_option = click.option(
    "-o", "--output", type=click.Choice(["text", "json", "tsv"]), default="text", help="Use alternative output format"
)
region_option = click.option(
    "--region",
    envvar="AWS_DEFAULT_REGION",
    metavar="AWS_REGION_ID",
    default=piu.utils.current_region(),
    help="AWS region ID (e.g. eu-central-1)",
    callback=piu.utils.validate_region,
)
odd_host_option = click.option(
    "-O", "--odd-host", help="Odd SSH bastion hostname", envvar="ODD_HOST", metavar="HOSTNAME"
)


def parse_time(s: str) -> float:
    """
    >>> parse_time('2015-04-14T19:09:01.000Z') > 0
    True
    """
    try:
        utc = datetime.datetime.strptime(s, "%Y-%m-%dT%H:%M:%S.%fZ")
        ts = time.time()
        utc_offset = datetime.datetime.fromtimestamp(ts) - datetime.datetime.utcfromtimestamp(ts)
        local = utc + utc_offset
        return local.timestamp()
    except Exception as e:
        print(e)
        return None


class AliasedDefaultGroup(AliasedGroup):
    def resolve_command(self, ctx, args):
        cmd_name = args[0]
        cmd = AliasedGroup.get_command(self, ctx, cmd_name)
        if not cmd:
            cmd_name = DEFAULT_COMMAND
            cmd = AliasedGroup.get_command(self, ctx, cmd_name)
            new_args = args
        else:
            new_args = args[1:]
        return cmd_name, cmd, new_args


def load_config(path):
    try:
        with open(path, "rb") as fd:
            return yaml.safe_load(fd)
    except Exception:
        return {}


def store_config(config, path):
    dir_path = os.path.dirname(path)
    if dir_path:
        os.makedirs(dir_path, exist_ok=True)
    with open(path, "w") as fd:
        yaml.dump(config, fd)


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo("Piu {}".format(piu.__version__))
    ctx.exit()


def tunnel_validation(ctx, param, value):
    if value and not re.match("^[0-9]{1,5}:[0-9]{1,5}$", value):
        raise click.BadParameter("Tunnel needs to be in format localPort:remotePort")
    else:
        return value


def lookup_instance(region, ip_address):
    filters = [{"Name": "network-interface.addresses.private-ip-address", "Values": [str(ip_address)]}]
    return next(piu.utils.list_running_instances(region, filters), None)


def ssh_keys_added():
    try:
        # ssh-add -l exits with status 1 (and check_call throws) if there are no keys in the SSH agent
        subprocess.check_call(["ssh-add", "-l"], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
        return True
    except Exception:
        return False


def _request_eic_access(
    ec2, hostname: str, remote_host: str, clip: bool, connect: bool, tunnel: bool, ssh_public_key: str
) -> bool:
    host_via = hostname
    click.secho("Requesting access with EIC to host {host_via}".format(host_via=host_via), bold=True)
    if remote_host:
        try:
            remote_attributes = instance_attributes(ec2, "private-ip-address", remote_host)
        except RuntimeError:
            print("Failed to get attributes for instances with private IP address {}".format(remote_host))
            return False
        if not send_ssh_key("ubuntu", remote_attributes, ssh_public_key):
            return False
    click.secho("Login credentials are valid only for 60 seconds. Please be quick.")
    ssh_connection(clip, connect, hostname, remote_host, tunnel, "odd", "ubuntu")
    return True


def _request_even_access(even_url, cacert, username, hostname, reason, remote_host, lifetime, clip, connect, tunnel):
    data = {"username": username, "hostname": hostname, "reason": reason}
    host_via = hostname
    if remote_host:
        data["remote_host"] = remote_host
        host_via = "{} via {}".format(remote_host, hostname)
    if lifetime:
        data["lifetime_minutes"] = lifetime
    try:
        access_token = zign.api.get_token("piu", ["uid"])
    except zign.api.ServerError as e:
        click.secho("{}".format(e), fg="red", bold=True)
        return False

    click.secho(
        "Requesting access with Even to host {host_via} for {username}..".format(host_via=host_via, username=username),
        bold=True,
    )
    r = requests.post(
        even_url,
        headers={"Content-Type": "application/json", "Authorization": "Bearer {}".format(access_token)},
        data=json.dumps(data),
        verify=cacert,
    )
    if r.status_code == 200:
        click.secho(r.text, fg="green", bold=True)
        ssh_connection(clip, connect, hostname, remote_host, tunnel, username, username)
    else:
        click.secho(
            "Server returned status {code}: {text}".format(code=r.status_code, text=r.text), fg="red", bold=True
        )
    return r.status_code == 200


def ssh_connection(clip, connect, hostname, remote_host, tunnel, odd_user, remote_user):
    if tunnel:
        ports = tunnel.split(":")
        ssh_command = "ssh {odd_user}@{hostname} -L {local_port}:{remote_host}:{remote_port}".format(
            odd_user=odd_user, hostname=hostname, local_port=ports[0], remote_host=remote_host, remote_port=ports[1]
        )
    else:
        if remote_host:
            ssh_command = "ssh -J {odd_user}@{odd_host} {remote_user}@{remote_host}".format(
                odd_user=odd_user, odd_host=hostname, remote_user=remote_user, remote_host=remote_host
            )
        else:
            ssh_command = "ssh {odd_user}@{odd_host}".format(odd_user=odd_user, odd_host=hostname)
    if connect or tunnel:
        subprocess.call(ssh_command.split())
    if not ssh_keys_added():
        warning("No SSH identities found. Please add one using ssh-add, for example:")
        warning("ssh-add ~/.ssh/id_rsa")
    click.secho("You can access your server with the following command:")
    click.secho(ssh_command)
    if clip:
        click.secho('\nOr just check your clipboard and run ctrl/command + v (requires package "xclip" on Linux)')
        if pyperclip is not None:
            pyperclip.copy(ssh_command)


@click.group(cls=AliasedDefaultGroup, context_settings=CONTEXT_SETTINGS)
@click.option(
    "--config-file", "-c", help="Use alternative configuration file", default=CONFIG_FILE_PATH, metavar="PATH"
)
@click.option(
    "-V",
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help="Print the current version number and exit.",
)
@click.pass_context
def cli(ctx, config_file):
    ctx.obj = config_file


@cli.command("request-access")
@click.argument("host", metavar="[USER]@HOST", required=False)
@click.argument("reason", required=False)
@click.argument("reason_cont", nargs=-1, metavar="[..]", required=False)
@click.option("-E", "--even-url", help="Even SSH Access Granting Service URL", envvar="EVEN_URL", metavar="URI")
@odd_host_option
@click.option(
    "-t",
    "--lifetime",
    help="Lifetime of the SSH access request in minutes (default: 60)",
    type=click.IntRange(1, 525600, clamp=True),
)
@click.option("--interactive", help="Offers assistance", envvar="PIU_INTERACTIVE", is_flag=True, default=False)
@click.option("--insecure", help="Do not verify SSL certificate", is_flag=True, default=False)
@click.option("--clip", help="Copy SSH command into clipboard", is_flag=True, default=False)
@click.option("--connect", help="Directly connect to the host", envvar="PIU_CONNECT", is_flag=True, default=False)
@click.option(
    "--tunnel",
    help="Tunnel to the host",
    envvar="PIU_TUNNEL",
    callback=tunnel_validation,
    metavar="LOCALPORT:REMOTEPORT",
)
@click.option(
    "--check/--no-check",
    help="Verify that the EC2 instance exists and wasn't shutdown",
    envvar="PIU_CHECK_INSTANCE",
    default=True,
)
@click.option(
    "-i",
    "--ssh-public-key",
    help="The public key to use to SSH",
    type=click.Path(),
    default=os.path.expanduser("~/.ssh/id_rsa.pub"),
)
@region_option
@click.pass_obj
def request_access(
    config_file,
    host,
    reason,
    reason_cont,
    even_url,
    odd_host,
    lifetime,
    interactive,
    insecure,
    clip,
    connect,
    tunnel,
    region,
    check,
    ssh_public_key,
):
    config = load_config(config_file)
    even_url = even_url or config.get("even_url")
    odd_host = odd_host or piu.utils.find_odd_host(region) or config.get("odd_host")
    if not check_ssh_key(ssh_public_key) and check_ssh_key(config.get("ssh_public_key")):
        ssh_public_key = config.get("ssh_public_key")

    if interactive:
        host, odd_host, reason, ssh_public_key = request_access_interactive(region, odd_host, ssh_public_key)
    if not host:
        raise click.UsageError('Missing argument "host".')
    if not reason:
        raise click.UsageError('Missing argument "reason".')

    if connect and tunnel:
        raise click.UsageError('Cannot specify both "connect" and "tunnel"')

    parts = host.split("@")
    if len(parts) > 1:
        username = parts[0]
    else:
        username = zign.api.get_config().get("user") or os.getenv("USER")

    hostname = parts[-1]

    try:
        ip = ipaddress.ip_address(hostname)

        if check and not interactive and ip in STUPS_CIDR:
            instance = lookup_instance(region, ip)
            if instance is None:
                click.confirm(
                    "No running instances found for {}, do you still want to request access?".format(ip), abort=True
                )
    except ValueError:
        ip = None

    reason = " ".join([reason] + list(reason_cont)).strip()

    cacert = not insecure
    if "cacert" in config:
        cacert = config["cacert"]

    while not even_url:
        even_url = click.prompt("Please enter the Even SSH access granting service URL")
        if not even_url.startswith("http"):
            # convenience for humans: add HTTPS by default
            even_url = "https://{}".format(even_url)
        try:
            requests.get(even_url)
        except Exception:
            error("Could not reach {}".format(even_url))
            even_url = None
        config["even_url"] = even_url

    while ip and ip in STUPS_CIDR and not odd_host:
        odd_host = click.prompt("Please enter the Odd SSH bastion hostname")
        try:
            socket.getaddrinfo(odd_host, 22)
        except Exception:
            error("Could not resolve hostname {}".format(odd_host))
            odd_host = None
        config["odd_host"] = odd_host

    while not check_ssh_key(ssh_public_key):
        ssh_public_key = click.prompt("Please enter path of a valid SSH public key")

    config["ssh_public_key"] = ssh_public_key
    store_config(config, config_file)

    if not even_url.endswith("/access-requests"):
        even_url = even_url.rstrip("/") + "/access-requests"

    first_host = hostname
    remote_host = hostname
    if odd_host:
        first_host = odd_host

    if first_host == remote_host:
        # user friendly behavior: it makes no sense to jump from bastion to itself
        remote_host = None
    elif remote_host.startswith("odd-"):
        # user friendly behavior: if the remote host is obviously a odd host, just use it
        first_host = remote_host
        remote_host = None

    ec2 = boto3.client("ec2")
    if not send_odd_ssh_key(ec2, first_host, ssh_public_key):
        error("Failed to send SSH key to odd host {host:s}".format(host=first_host))
        sys.exit(1)

    use_eic = True
    if remote_host:
        try:
            remote_attributes = instance_attributes(ec2, "private-ip-address", remote_host)
        except RuntimeError:
            print("Failed to find instance with IP {0:s}".format(remote_host))
            sys.exit(1)
        use_eic = compatible_ami(ec2, remote_attributes["ImageId"])

    if use_eic:
        success = _request_eic_access(ec2, first_host, remote_host, clip, connect, tunnel, ssh_public_key)
    else:
        success = _request_even_access(
            even_url, cacert, username, first_host, reason, remote_host, lifetime, clip, connect, tunnel
        )
    if not success:
        sys.exit(1)


def check_ssh_key(key_path: str) -> bool:
    # TODO: verify that the input key is actually an SSH public key
    if key_path and os.path.exists(key_path):
        return True
    return False


def send_odd_ssh_key(ec2, odd_hostname: str, public_key: str) -> bool:
    odd_ip = socket.gethostbyname(odd_hostname)
    try:
        odd_attributes = instance_attributes(ec2, "ip-address", odd_ip)
    except RuntimeError as e:
        print("Failed to find odd host {0:s}: {1:s}".format(odd_hostname), e)
    return send_ssh_key("odd", odd_attributes, public_key)


def send_ssh_key(username: str, host: dict, public_key: str) -> bool:
    with open(public_key) as key:
        contents = key.read()
    eic = boto3.client("ec2-instance-connect")
    result = eic.send_ssh_public_key(
        InstanceId=host["InstanceId"],
        InstanceOSUser=username,
        SSHPublicKey=contents,
        AvailabilityZone=host["Placement"]["AvailabilityZone"],
    )
    return result["Success"]


def request_access_interactive(region, odd_host, ssh_public_key):
    region = click.prompt("AWS region", default=region)
    odd_host = click.prompt("Odd SSH bastion hostname", default=odd_host)

    all_instances = piu.utils.list_running_instances(region, [])

    stack_instances = [
        instance for instance in all_instances if instance.name and instance.stack_name and instance.stack_version
    ]

    instance_count = len(stack_instances)
    if instance_count == 0:
        raise click.ClickException("No running instances were found.")

    stack_instances.sort(key=operator.attrgetter("stack_name", "stack_version"))

    print()
    table_entries = [dict(index=idx, **instance._asdict()) for idx, instance in enumerate(stack_instances, start=1)]
    print_table("index name stack_name stack_version private_ip instance_id".split(), table_entries)
    print()

    if instance_count > 1:
        allowed_choices = ["{}".format(n) for n in range(1, instance_count + 1)]
        instance_index = (
            int(click.prompt("Choose an instance (1-{})".format(instance_count), type=click.Choice(allowed_choices)))
            - 1
        )
    else:
        click.confirm("Connect to {}?".format(stack_instances[0].name), default=True, abort=True)
        instance_index = 0

    host = stack_instances[instance_index].private_ip
    reason = click.prompt("Reason", default="Troubleshooting")
    ssh_public_key = click.prompt("SSH Public Key", default=ssh_public_key)
    return host, odd_host, reason, ssh_public_key


@cli.command("list-access-requests")
@click.option("-u", "--user", help="Filter by username", metavar="NAME")
@odd_host_option
@click.option("-s", "--status", help="Filter by status", metavar="NAME", type=click.Choice(STATUS_NAMES))
@click.option("-l", "--limit", help="Limit number of results", type=int, default=20)
@click.option("--offset", help="Offset", type=int, default=0)
@output_option
@region_option
@click.pass_obj
def list_access_requests(config_file, user, odd_host, status, limit, offset, output, region):
    """List access requests filtered by user, host and status"""
    config = load_config(config_file)

    if user == "*":
        user = None

    if odd_host == "*":
        odd_host = None
    elif odd_host is None:
        odd_host = piu.utils.find_odd_host(region) or config.get("odd_host")

    access_token = zign.api.get_token("piu", ["piu"])

    params = {"username": user, "hostname": odd_host, "status": status, "limit": limit, "offset": offset}
    r = requests.get(
        config.get("even_url").rstrip("/") + "/access-requests",
        params=params,
        headers={"Authorization": "Bearer {}".format(access_token)},
    )
    r.raise_for_status()
    rows = []
    for req in r.json():
        req["created_time"] = parse_time(req["created"])
        rows.append(req)
    rows.sort(key=lambda x: x["created_time"])
    with OutputFormat(output):
        print_table(
            "username hostname remote_host reason lifetime_minutes status status_reason created_time".split(),
            rows,
            styles=STYLES,
            titles=TITLES,
            max_column_widths=MAX_COLUMN_WIDTHS,
        )


def instance_attributes(ec2, filter_name: str, filter_value: str) -> dict:
    instances = ec2.describe_instances(Filters=[{"Name": filter_name, "Values": [filter_value]}])
    if len(instances["Reservations"]) < 1 or len(instances["Reservations"][0]["Instances"]) < 1:
        raise RuntimeError("Failed to find instance with {0:s}: {1:s}".format(filter_name, filter_value))
    return instances["Reservations"][0]["Instances"][0]


def compatible_ami(ec2, ami_id: str) -> bool:
    amis = ec2.describe_images(ImageIds=[ami_id])
    if len(amis["Images"]) != 1:
        raise RuntimeError("Failed to determine AMI properties for ami {0:s}".format(ami_id))
    ami_creation = dateutil.parser.parse(amis["Images"][0]["CreationDate"])
    if ami_creation > dateutil.parser.parse(COMPATIBLE_AMI_CUTOFF):
        return True
    return False


def main():
    handle_exceptions(cli)()


if __name__ == "__main__":
    main()
